#!/usr/bin/env python
# coding: utf-8

import re
import os

#  https://stackoverflow.com/questions/31801271/what-are-the-supported-git-url-formats
#  I found the list below. It is not complete since
#  https://<token>:x-oauth-basic@host.xz/path/to/repo.git is not there.
#
#  Secure Shell Transport Protocol
#  ssh://user@host.xz:port/path/to/repo.git/
#  ssh://user@host.xz/path/to/repo.git/
#  ssh://host.xz:port/path/to/repo.git/
#  ssh://host.xz/path/to/repo.git/
#  ssh://user@host.xz/path/to/repo.git/
#  ssh://host.xz/path/to/repo.git/
#  ssh://user@host.xz/~user/path/to/repo.git/
#  ssh://host.xz/~user/path/to/repo.git/
#  ssh://user@host.xz/~/path/to/repo.git
#  ssh://host.xz/~/path/to/repo.git
#  user@host.xz:/path/to/repo.git/
#  host.xz:/path/to/repo.git/
#  user@host.xz:~user/path/to/repo.git/
#  host.xz:~user/path/to/repo.git/
#  user@host.xz:path/to/repo.git
#  host.xz:path/to/repo.git
#  rsync://host.xz/path/to/repo.git/
#
#  Git Transport Protocol
#  git://host.xz/path/to/repo.git/
#  git://host.xz/~user/path/to/repo.git/
#
#  HTTP/S Transport Protocol
#  http://host.xz/path/to/repo.git/
#  https://host.xz/path/to/repo.git/

#  Secure Shell Transport Protocol

#  ssh://(?P<user>.*?)@(?P<host>.*?):(?P<port>.*?)/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  ssh://(?P<user>.*?)@(?P<host>.*?)/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  ssh://(?P<host>.*?):(?P<port>.*?)/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  ssh://(?P<host>.*?)/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  ssh://(?P<user>.*?)@(?P<host>.*?)/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  ssh://(?P<host>.*?)/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  ssh://(?P<user>.*?)@(?P<host>.*?)/~user/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  ssh://(?P<host>.*?)/~(?P<user>.*?)/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  ssh://(?P<user>.*?)@(?P<host>.*?)/~/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  ssh://(?P<host>.*?)/~/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  (?P<user>.*?)@(?P<host>.*?):/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  (?P<host>.*?):/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  (?P<user>.*?)@(?P<host>.*?):~user/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  (?P<host>.*?):~(?P<user>.*?)/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  (?P<user>.*?)@(?P<host>.*?):(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  (?P<host>.*?):(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  rsync://(?P<host>.*?)/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$

#  Git Transport Protocol

#  git://(?P<host>.*?)/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  git://(?P<host>.*?)/~(?P<user>.*?)/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$

#  HTTP/S Transport Protocol

#  http://(?P<host>.*?)/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$
#  https://(?P<host>.*?)/(?P<user>.*?)/(?P<repo>.*?)(\.git)?(/)?(?P<branch>@.*?)?$


rule_groups = [{
    "defaults": {
        "host": "github.com",
    },
    "env": {
        "committer": "GITHUB_USERNAME",
        "token": "GITHUB_TOKEN",
    },
    "provider": "github.com",
    "fmt": {
        "ssh": 'git@{host}:{user}/{repo}.git',
        "https": 'https://{host}/{user}/{repo}.git',
        "https_token": 'https://{committer}:{token}@{host}/{user}/{repo}.git',
    },
    "patterns": [
        # github.com/openacid/slim.git
        ('ssh',  r'github.com/(?P<user>.+?)/(?P<repo>.+?)(\.git)?/?(?P<branch>@.+?)?$'),
        # git@github.com:openacid/slim.git
        ('ssh', r'git@github.com:(?P<user>.+?)/(?P<repo>.+?)(\.git)?/?(?P<branch>@.+?)?$'),
        # ssh://git@github.com/openacid/openacid.github.io
        ('ssh', r'ssh://git@github.com/(?P<user>.+?)/(?P<repo>.+?)(\.git)?/?(?P<branch>@.+?)?$'),
        # https://committer:token@github.com/openacid/openacid.github.io.git
        ('https', r'https://(?P<committer>.+?):(?P<token>.+?)@github.com/(?P<user>.+?)/(?P<repo>.+?)(\.git)?/?(?P<branch>@.+?)?$'),
        # http://github.com/openacid/openacid.github.io.git
        ('https', r'http://github.com/(?P<user>.+?)/(?P<repo>.+?)(\.git)?/?(?P<branch>@.+?)?$'),
        # https://github.com/openacid/openacid.github.io.git
        ('https', r'https://github.com/(?P<user>.+?)/(?P<repo>.+?)(\.git)?/?(?P<branch>@.+?)?$'),
    ],
}, {
    "defaults": {
    },
    "env": {
    },
    "provider": "*",
    "fmt": {
        "ssh": 'git@{host}:{user}/{repo}.git',
        "https": 'https://{host}/{user}/{repo}.git',
        "https_token": 'https://{committer}:{token}@{host}/{user}/{repo}.git',
    },
    "patterns": [
        ('https', r'https://(?P<committer>.+?):(?P<token>.+?)@(?P<host>.+?)/(?P<user>.+?)/(?P<repo>.+?)(\.git)?/?(?P<branch>@.+?)?$'),
        ('https', r'http://(?P<host>[^/]+?)/(?P<user>.+?)/(?P<repo>.+?)(\.git)?/?(?P<branch>@.+?)?$'),
        ('https', r'https://(?P<host>[^/]+?)/(?P<user>.+?)/(?P<repo>.+?)(\.git)?/?(?P<branch>@.+?)?$'),
        ('ssh',   r'ssh://git@(?P<host>[^/]+?)/(?P<user>.+?)/(?P<repo>.+?)(\.git)?/?(?P<branch>@.+?)?$'),
        ('ssh',   r'git@(?P<host>[^/]+?):(?P<user>.+?)/(?P<repo>.+?)(\.git)?/?(?P<branch>@.+?)?$'),
        ('ssh',   r'(?P<host>[^/]+?)/(?P<user>.+?)/(?P<repo>.+?)(\.git)?/?(?P<branch>@.+?)?$'),
    ],

},
]


class GitUrl(object):
    """
    GitUrl parse and format git urls
    """

    def __init__(self, fields, rule_group, matching_pattern):
        """
        Create a GitUrl object.

        Args:

            fields(dict): fields of url components, such as 'user', 'repo',
                    'branch', 'token', 'committer'.

            rule_group(dict): one of the predefined rule group this git-url
                    matched and is also used to output plain text url.

            matching_pattern(str): the url pattern defined in rule_group that matches this url.
        """

        self.fields = fields
        self.rule_group = rule_group
        self.matching_pattern = matching_pattern

    def fmt(self, scheme=None):
        """
        format git url to scheme ssh or https

        Args:

           scheme(str): specifies the output url format:

                        - ``"ssh": 'git@{host}:{user}/{repo}.git'``,

                        - ``"https": 'https://{host}/{user}/{repo}.git'``,

                        - ``"https" with token present in fields: 'https://{committer}:{token}@{host}/{user}/{repo}.git'``,

                        If absent, format by fields['sheme']

        Returns:
            str: the formatted url
        """

        if scheme is None:
            scheme = self.fields['scheme']

        if scheme == 'https':
            if 'token' in self.fields:
                fmt = self.rule_group['fmt']['https_token']
            else:
                fmt = self.rule_group['fmt']['https']
        elif scheme == 'ssh':
            fmt = self.rule_group['fmt']['ssh']
        else:
            raise ValueError("invalid scheme: "+scheme)

        return fmt.format(**self.fields)

    @classmethod
    def parse(cls, url):
        """
        Parse plain text git url and return an instance of GitUrl.

        Args:

            url(str): git url in string in one form of:

                    - ``git@github.com:openacid/slim.git``
                    - ``ssh://git@github.com/openacid/openacid.github.io``
                    - ``https://committer:token@github.com/openacid/openacid.github.io.git``
                    - ``http://github.com/openacid/openacid.github.io.git``
                    - ``https://github.com/openacid/openacid.github.io.git``

        Returns:
            GitUrl
        """

        for g in rule_groups:
            for (scheme, p) in g['patterns']:
                match = re.match(p, url)
                if not match:
                    continue

                d = match.groupdict()
                d.update(g['defaults'])

                d['scheme'] = scheme

                #  extend vars from env
                for var_name, env_name in g['env'].items():
                    if var_name not in d:
                        v = os.environ.get(env_name)
                        if v is not None:
                            d[var_name] = v

                return cls(d, g, p)

        raise ValueError(
            'unknown url: {url};'.format(
                url=url,
                )
        )



"""
k3git is wrapper of git command-line.

To parse a git command ``git --git-dir=/foo fetch origin``:

    >>> GitOpt().parse_args(['--git-dir=/foo', 'fetch', 'origin']).cmds
    ['fetch', 'origin']

    >>> GitOpt().parse_args(['--git-dir=/foo', 'fetch', 'origin']).to_args()
    ['--git-dir=/foo']

"""

__version__ = "0.1.6"
__name__ = "k3git"

from .gitopt import GitOpt
from .giturl import GitUrl
from .git_wrapper import Git

__all__ = [
    'Git',
    'GitOpt',
]



#!/usr/bin/env python
# coding: utf-8

import os
import logging


from k3str import to_utf8
from k3handy import pabs

from k3handy import cmdf

logger = logging.getLogger(__name__)


class Git(object):

    def __init__(self, opt, gitpath=None, gitdir=None, working_dir=None, cwd=None, ctxmsg=None):
        self.opt = opt.clone()
        # gitdir and working_dir is specified and do not consider '-C' option
        if gitdir is not None:
            self.opt.opt['git_dir'] = pabs(gitdir)
        if working_dir is not None:
            self.opt.opt['work_tree'] = pabs(working_dir)

        self.cwd = cwd

        self.gitpath = gitpath or "git"
        self.ctxmsg = ctxmsg

    # high level API

    def checkout(self, branch, flag='x'):
        self.cmdf("checkout", branch, flag=flag)

    def fetch(self, name, flag=''):
        self.cmdf("fetch", name, flag=flag)

    # branch

    def branch_default_remote(self, branch, flag=''):
        """
        Returns the default remote name of a branch.
        """
        return self.cmdf('config', '--get',
                    'branch.{}.remote'.format(branch),
                    flag=flag+'n0')

    # head

    def head_branch(self, flag=''):
        """
        Returns the branch HEAD pointing to.
        """
        return self.cmdf('symbolic-ref', '--short', 'HEAD', flag=flag+'n0')


    # remote

    def remote_get(self, name, flag=''):
        return self.cmdf("remote", "get-url", name, flag=flag + 'n0')

    def remote_add(self, name, url, flag='x', **options):
        self.cmdf("remote", "add", name, url, **options, flag=flag)

    # blob

    def blob_new(self, f, flag=''):
        return self.cmdf("hash-object", "-w", f, flag=flag + 'n0')

    #  tree

    def tree_of(self, commit, flag=''):
        return self.cmdf("rev-parse", commit + "^{tree}", flag=flag + 'n0')

    def tree_items(self, treeish, name_only=False, with_size=False, flag='x'):
        args = []
        if name_only:
            args.append("--name-only")

        if with_size:
            args.append("--long")
        return self.cmdf("ls-tree", treeish, *args, flag=flag + 'no')

    def tree_add_obj(self, cur_tree, path, treeish):

        sep = os.path.sep

        itms = self.tree_items(cur_tree)

        if sep not in path:
            return self.tree_new(itms, path, treeish, flag='x')

        # a/b/c -> a, b/c
        p0, left = path.split(sep, 1)
        p0item = self.tree_find_item(cur_tree, fn=p0, typ="tree")

        if p0item is None:

            newsubtree = treeish
            for p in reversed(left.split(sep)):
                newsubtree = self.tree_new([], p, newsubtree, flag='x')
        else:

            subtree = p0item["object"]
            newsubtree = self.tree_add_obj(subtree, left, treeish)

        return self.tree_new(itms, p0, newsubtree, flag='x')

    def tree_find_item(self, treeish, fn=None, typ=None):
        for itm in self.tree_items(treeish):
            itm = self.parse_tree_item(itm)
            if fn is not None and itm["fn"] != fn:
                continue
            if typ is not None and itm["type"] != typ:
                continue

            return itm
        return None

    def parse_tree_item(self, line):

        # git-ls-tree output:
        #     <mode> SP <type> SP <object> TAB <file>
        # This output format is compatible with what --index-info --stdin of git update-index expects.
        # When the -l option is used, format changes to
        #     <mode> SP <type> SP <object> SP <object size> TAB <file>
        # E.g.:
        # 100644 blob a668431ae444a5b68953dc61b4b3c30e066535a2    imsuperman
        # 040000 tree a668431ae444a5b68953dc61b4b3c30e066535a2    foo

        p, fn = line.split("\t", 1)

        elts = p.split()
        rst = {
            "mode": elts[0],
            "type": elts[1],
            "object": elts[2],
            "fn": fn,
        }
        if len(elts) == 4:
            rst["size"] = elts[3]

        return rst

    def tree_new(self, itms, name, obj, mode=None, flag='x'):

        newitems = [x for x in itms
                    if self.parse_tree_item(x)["fn"] != name]

        itm = self.treeitem_new(name, obj, mode=mode)

        newitems.append(itm)
        new_treeish = self.cmdf("mktree", input="\n".join(newitems), flag=flag + 'n0')
        return new_treeish

    # treeitem

    def treeitem_new(self, name, obj, mode=None):

        typ = self.obj_type(obj, flag='x')
        item_fmt = "{mode} {typ} {object}\t{name}"

        if typ == 'tree':
            mod = "040000"
        else:
            if mode is None:
                mod = "100644"
            else:
                mod = mode

        itm = item_fmt.format(mode=mod,
                              typ=typ,
                              object=obj,
                              name=name
                              )
        return itm

    # rev

    def rev_of(self, name, flag=''):
        """
        Get the revision(sha256) of an object.

        Args:

            name(str): could be short hash, full length hash, ref name or branch.

            flag(str): flag='x' to raise if return code is not 0. flag='' to return None.

        Returns:
            str: sha256 in lower-case hex. If no such object is found, it returns None.
        """
        return self.cmdf("rev-parse", "--verify", "--quiet", name, flag=flag + 'n0')

    def obj_type(self, obj, flag=''):
        return self.cmdf("cat-file", "-t", obj, flag=flag + 'n0')

    # wrapper of cli

    def _opt(self, **kwargs):
        opt = {}
        if self.cwd is not None:
            opt["cwd"] = self.cwd
        opt.update(kwargs)
        return opt

    def _args(self):
        return self.opt.to_args()

    def cmdf(self, *args, flag='', **kwargs):
        return cmdf(self.gitpath, *self._args(), *args, flag=flag, **self._opt(**kwargs))

    def out(self, fd, *msg):

        if self.ctxmsg is not None:
            os.write(fd, to_utf8(self.ctxmsg) + b": ")

        for (i, m) in enumerate(msg):
            os.write(fd, to_utf8(m))
            if i != len(msg) - 1:
                os.write(fd, b" ")
        os.write(fd, b"\n")



#!/usr/bin/env python
# coding: utf-8

import copy
import logging

logger = logging.getLogger(__name__)


class GitOpt(object):
    """
    GitOpt parses and builds git command line arguments.

    Attributes:

        cmds: parsed command, e.g. the cmds of parsed ``git --git-dir=/foo fetch origin`` is ``['fetch', 'origin']``.

        opt: parsed options.

        informative_cmds: defines the git options that is a query-like command,
            such as ``--version`` or ``--help``.

        additional: GitOpt is able to parse user defined options::

            o = GitOpt().parse_args(['--foo', 'fetch'], additional=['--foo'])
            o.additional
            # {'--foo': '--foo'}
    """

    # informative options just query for some info instead of doing anything.
    informative_opts = (
        '--version',
        '--help',
        '--html-path',
        '--info-path',
        '--man-path',
        '--exec-path',
    )

    def __init__(self):
        self.opt = {
            "startpath": [],
            "confkv": [],
            "paging": None,
            "no_replace_objects": False,
            "bare": False,
            "git_dir": None,
            "work_tree": None,
            "namespace": None,
            "super_prefix": None,
            "exec_path": None,
        }
        self.informative_cmds = {}
        self.additional = {}

        self.cmds = []

    def update(self, d):
        """
        update ``opt`` with a dictionary ``d``.

        Returns:
            self
        """
        for k, v in d.items():
            self.opt[k] = v
        return self

    def clone(self):
        """
        clone a GitOpt object so that the returned object share nothing with the original.

        Returns:
            GitOpt: a same and standalone object.
        """
        o = GitOpt()
        o.opt = copy.deepcopy(self.opt)
        o.informative_cmds = copy.deepcopy(self.informative_cmds)
        o.additional = copy.deepcopy(self.additional)
        return o

    def parse_args(self, args, additional=None):
        """
        Parse a command line input(without the "git").
        Additional user defined arguments can be specified.

        Returns:
            self
        """
        while len(args) > 0:
            arg = args.pop(0)

            if arg in self.informative_opts:
                self.informative_cmds[arg] = arg
                continue

            if arg == '-C':
                self.opt["startpath"].append(args.pop(0))
                continue
            if arg == '-c':
                self.opt["confkv"].append(args.pop(0))
                continue
            if arg.startswith('--exec-path='):
                self.opt["exec_path"] = arg.split('=', 1)[1]
                continue
            if arg in ('-p', '--paginate'):
                self.opt["paging"] = True
                continue
            if arg == '--no-pager':
                self.opt["paging"] = False
                continue

            if arg == '--no-replace-objects':
                # TODO
                self.opt["no_replace_objects"] = True
                continue

            if arg == '--bare':
                # TODO
                self.opt["bare"] = True
                continue

            if arg.startswith('--git-dir='):
                self.opt["git_dir"] = arg.split('=', 1)[1]
                continue

            if arg.startswith('--work-tree='):
                self.opt["work_tree"] = arg.split('=', 1)[1]
                continue

            if arg.startswith('--namespace='):
                # TODO
                self.opt["namespace"] = arg.split('=', 1)[1]
                continue
            if arg.startswith('--super-prefix='):
                # TODO
                self.opt["super_prefix"] = arg.split('=', 1)[1]
                continue

            if additional is not None:
                if arg in additional:
                    self.additional[arg] = arg
                    continue

            # no match, push back
            self.cmds = [arg] + args
            break

        return self

    def to_args(self):
        """
        Build git command line argument.
        E.g.::

            o = GitOpt().parse_args(['--git-dir=/foo', 'fetch'])
            o.opt['work_tree'] = '/bar'

            o.to_args() # ['--git-dir=/foo', '--work-tree=/bar']
            o.cmds      # ['fetch']

        Returns:
            list: of str that can be used in commandline.
        """
        o = self.opt
        rst = []
        for p in o['startpath']:
            rst.append('-C')
            rst.append(p)

        for kv in o['confkv']:
            rst.append('-c')
            rst.append(kv)

        if o['exec_path'] is not None:
            rst.append('--exec-path=' + o['exec_path'])

        if o['paging'] is True:
            rst.append('-p')

        if o['paging'] is False:
            rst.append('--no-pager')

        if o['no_replace_objects'] is True:
            rst.append('--no-replace-objects')

        if o['bare'] is True:
            rst.append('--bare')

        if o['git_dir'] is not None:
            rst.append('--git-dir=' + o['git_dir'])

        if o['work_tree'] is not None:
            rst.append('--work-tree=' + o['work_tree'])

        if o['namespace'] is not None:
            rst.append('--namespace=' + o['namespace'])

        if o['super_prefix'] is not None:
            rst.append('--super-prefix=' + o['super_prefix'])

        return rst



# DO NOT EDIT!!! built with `python _building/build_setup.py`
import setuptools
setuptools.setup(
    name="k3git",
    packages=["k3git"],
    version="0.1.6",
    license='MIT',
    description='wrapper of git command-line',
    long_description="# k3git\n\n[![Build Status](https://travis-ci.com/pykit3/k3git.svg?branch=master)](https://travis-ci.com/pykit3/k3git)\n![Python package](https://github.com/pykit3/k3git/workflows/Python%20package/badge.svg)\n[![Documentation Status](https://readthedocs.org/projects/k3git/badge/?version=stable)](https://k3git.readthedocs.io/en/stable/?badge=stable)\n[![Package](https://img.shields.io/pypi/pyversions/k3git)](https://pypi.org/project/k3git)\n\nwrapper of git command-line\n\nk3git is a component of [pykit3] project: a python3 toolkit set.\n\n\nk3git is wrapper of git command-line.\n\nTo parse a git command ``git --git-dir=/foo fetch origin``:\n\n    >>> GitOpt().parse_args(['--git-dir=/foo', 'fetch', 'origin']).cmds\n    ['fetch', 'origin']\n\n    >>> GitOpt().parse_args(['--git-dir=/foo', 'fetch', 'origin']).to_args()\n    ['--git-dir=/foo']\n\n\n\n\n# Install\n\n```\npip install k3git\n```\n\n# Synopsis\n\n```python\n>>> GitOpt().parse_args(['--git-dir=/foo', 'fetch', 'origin']).cmds\n['fetch', 'origin']\n>>> GitOpt().parse_args(['--git-dir=/foo', 'fetch', 'origin']).to_args()\n['--git-dir=/foo']\n```\n\n#   Author\n\nZhang Yanpo (张炎泼) <drdr.xp@gmail.com>\n\n#   Copyright and License\n\nThe MIT License (MIT)\n\nCopyright (c) 2015 Zhang Yanpo (张炎泼) <drdr.xp@gmail.com>\n\n\n[pykit3]: https://github.com/pykit3",
    long_description_content_type="text/markdown",
    author='Zhang Yanpo',
    author_email='drdr.xp@gmail.com',
    url='https://github.com/pykit3/k3git',
    keywords=['git', 'cli', 'commandline'],
    python_requires='>=3.0',

    install_requires=['k3ut<0.2,>=0.1.7', 'k3handy>=0.1.4,<0.2', 'k3fs<0.2,>=0.1.5'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
    ] + ['Programming Language :: Python :: 3.6', 'Programming Language :: Python :: 3.7', 'Programming Language :: Python :: 3.8', 'Programming Language :: Python :: Implementation :: PyPy'],
)
